using System;
using UnityEngine;
using UnityEngine.Experimental.Rendering;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using SimpleJSON;
using System.Linq;
using MVR;
using MeshVR;
using MVR.FileManagementSecure;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using GPUTools.Cloth.Scripts;
using Unity.Collections;
using AssetBundles;
using System.Diagnostics;
using MVR.FileManagementSecure;

namespace Stopper {
	public class XrayBody : MVRScript {
		
		GameObject extraCameraContainer;
		GameObject meshContainer;
		MeshRenderer renderer;
		Camera camera;
		
		// Layer to use as a culling mask for the extra camera
		// 18 appears unused 
		static int layer = 18; 
		static int cullingMask = 1 << layer;
		
		DAZMergedSkinV2 skin;
		
		Material[] oldMaterials;
		
		GameObject materialUIPrefab;
		MVRScriptUI scriptUI;
		
		MaterialConfig[] materialConfigs;
		
		bool isPerson = true;
		Dictionary<Renderer, int> originalRendererLayers = new Dictionary<Renderer, int>();
		
		public class MaterialConfig {
			public string name;
			public int index;
			
			public Material material;
			
			public JSONStorableBool enabled;
			public JSONStorableBool cull;	
			public UIDynamicButton materialOptionsButton;
			
			public MaterialOptionsWrapper matOptions;
			public Transform matOptionsUI;
		}
		
		public class MaterialOptionsWrapper : MaterialOptions {
			public void SetParams(){
				this.SetAllParameters();
			}
		}
		
		public override void Init() {
			
			
			// Add a container for the extra camera that'll draw on top of the main camera
			// This will be parented off of the main camera so that it follows it
			extraCameraContainer = GetOrCreateChildObject(Camera.main.gameObject, "extraCamera");
			extraCameraContainer.transform.parent = Camera.main.gameObject.transform;
			
			
			camera = extraCameraContainer.GetComponent<Camera>();
			if(camera == null){
				camera = extraCameraContainer.AddComponent<Camera>();
			}
			
			camera.CopyFrom(Camera.main);
			camera.clearFlags = CameraClearFlags.Depth;
			camera.depth = Camera.main.depth +1;
			camera.cullingMask = cullingMask;
			
			
			skin = containingAtom.GetComponentInChildren<DAZMergedSkinV2>();
			
			// Some basic support for non-person atoms. It moves the object to the overlay camera's
			// layer so that they get rendered on top, but doesn't do anything with the materials.
			if(skin == null){
				isPerson = false;
				foreach(Renderer renderer in containingAtom.GetComponentsInChildren<Renderer>()){
					originalRendererLayers.Add(renderer, renderer.gameObject.layer);
					renderer.gameObject.layer = layer;
				}
				return;
			}
			
			
			// Find a pre-existing material option UI to use as a template. 
			// Any would likely do...
			foreach (GameObject go in Resources.FindObjectsOfTypeAll(typeof(GameObject)) as GameObject[]){
				if(go.name =="F Eyelash Materials") {
					materialUIPrefab = go;
					break;
				}
			}
			
			// Create another container to hold the additional renderer and meshes for this atom
			meshContainer = GetOrCreateChildObject(containingAtom.gameObject, "extraMeshes");
			meshContainer.layer = layer; // This layer otherwise appears unused
		
			renderer = meshContainer.GetComponent<MeshRenderer>();
			if(renderer == null){
				renderer = meshContainer.AddComponent<MeshRenderer>();
			}
			
			MeshFilter filter = meshContainer.GetComponent<MeshFilter>();
			if(filter == null){
				filter = meshContainer.AddComponent<MeshFilter>();
			}
			
			
			
			filter.sharedMesh = skin.Mesh;
			
			// Save the original materials
			oldMaterials = skin.GPUmaterials;
			
			materialConfigs = new MaterialConfig[skin.GPUmaterials.Length];
			for(int i = 0; i < skin.GPUmaterials.Length; i++){
				Material mat = skin.GPUmaterials[i];
				
				MaterialConfig matConfig = new MaterialConfig();
				materialConfigs[i] = matConfig;
				
				matConfig.name = mat.name;
				matConfig.index = i;
				matConfig.enabled = new JSONStorableBool(mat.name, false);
				RegisterBool(matConfig.enabled);
				CreateToggle(matConfig.enabled, false);
				
				
				UIDynamicButton materialOptionsButton = CreateButton("Material Options");
				matConfig.materialOptionsButton = materialOptionsButton;
				materialOptionsButton.button.onClick.AddListener(() => {
					ShowMaterialOptionUI(matConfig.index);
                });
				
				
				if(matConfig.material == null){
					matConfig.material = new Material(skin.GPUmaterials[matConfig.index]);
					matConfig.material.shader = Shader.Find("Custom/Subsurface/TransparentGlossNMSeparateAlphaComputeBuff");
					matConfig.material.SetFloat("_AlphaAdjust", -0.3f);
				}
				
				bool createdMatOpts = false;
				if(matConfig.matOptions == null){
					createdMatOpts = true;
					matConfig.matOptions = meshContainer.AddComponent<MaterialOptionsWrapper>();
				}
				
				matConfig.matOptionsUI = UnityEngine.Object.Instantiate<Transform>(materialUIPrefab.transform);
				
				matConfig.matOptions.SetUI(matConfig.matOptionsUI);
				
				matConfig.matOptions.materialContainer = meshContainer.transform;
				matConfig.matOptions.textureGroup1 = new MaterialOptionTextureGroup();
				matConfig.matOptions.SetCustomTextureFolder("Custom/Atom/Person/Textures");
				
				
				matConfig.matOptions.materialForDefaults = matConfig.material;
		
				matConfig.matOptions.paramMaterialSlots = new int[]{matConfig.index};
				matConfig.matOptions.textureGroup1.materialSlots = matConfig.matOptions.paramMaterialSlots;

				matConfig.matOptions.SyncAllParameters();
				
				matConfig.matOptionsUI.gameObject.SetActive(false);
				
				
				
			}
		}
		
		private void Start(){
			if(!isPerson){
				return;
			}
			
			// The material selction prefab doesn't do well with a layout,
			// so we're just going to remove it.
			scriptUI = this.UITransform.GetComponentInChildren<MVRScriptUI>();
			
			UnityEngine.UI.VerticalLayoutGroup layout = scriptUI.fullWidthUIContent.gameObject.GetComponent<UnityEngine.UI.VerticalLayoutGroup>();
			Destroy(layout);
			
			// Add the back button 
			Transform backButtonTransform = UnityEngine.Object.Instantiate<Transform>(manager.configurableButtonPrefab.transform);
			backButtonTransform.SetParent(scriptUI.fullWidthUIContent,false);
			
			UIDynamicButton backButton = backButtonTransform.GetComponent<UIDynamicButton>();
			backButton.label = "BACK";
			backButton.buttonColor = Color.green;
			backButton.button.onClick.AddListener(() => {
				HideMaterialOptionUI();
			});
			
			// Without a layout these don't work predictably
			// These settings put the back button more or less out of the way
			// and I'm calling that good enough
			RectTransform rect = backButtonTransform.gameObject.GetComponent<RectTransform>();
			rect.offsetMax = new Vector2(0f,0f);
			rect.anchorMax = new Vector2(0.15f,1f);
			
			foreach(MaterialConfig matConfig in materialConfigs){
				matConfig.matOptionsUI.SetParent(scriptUI.fullWidthUIContent,false);
				
				if(matConfig.enabled.val){
					matConfig.materialOptionsButton.gameObject.SetActive(true);
				} else {
					matConfig.materialOptionsButton.gameObject.SetActive(false);
				}
				
				matConfig.enabled.toggle.onValueChanged.AddListener((val) => {
					SyncMaterials();
					if(val)
						matConfig.materialOptionsButton.gameObject.SetActive(true);
					else 
						matConfig.materialOptionsButton.gameObject.SetActive(false);
				});
			}
			
			scriptUI.fullWidthUIContent.gameObject.SetActive(false);
			
			SyncMaterials(true);
		}
		
		private void Update()
		{
			if(!isPerson){
				return;
			}
			
			/* 	We have to do a bit of a dance here...
				VAM's shaders require some buffers to be set in order to work properly.
				Since we're using copies of the original materials, these buffers don't 
				get set by default. DAZMergedSkinV2 sets these on its GPUMaterials when
				it draws, so we temporarily replace the skin's GPUMaterials with our set,
				wait for the buffers to set, and swap them back. 
				
				If it's stupid but works...
			*/
			if(oldMaterials != null && skin.GPUmaterials != oldMaterials){
				bool allBuffersSet = true;
				foreach(MaterialConfig matConfig in materialConfigs){
					if(!matConfig.enabled.val)
						continue;
					
					if(!matConfig.material.HasProperty("verts")){
						allBuffersSet = false;
					}
				}
				
				if(allBuffersSet){
					skin.GPUmaterials = oldMaterials;
					skin.FlushBuffers();
					
					foreach(MaterialConfig matConfig in materialConfigs){
						matConfig.matOptions.SetParams();
					}
				}
				return;
			} 
		}
		
		private void OnEnable(){
			if(renderer != null){
				renderer.enabled = true;
			}
			
			if(!isPerson){
				foreach(Renderer renderer in containingAtom.GetComponentsInChildren<Renderer>()){
					renderer.gameObject.layer = layer;
				}	
			}
		}
		
		private void OnDisable()
		{ 
			if(skin != null && oldMaterials != null){
				skin.GPUmaterials = oldMaterials;
				skin.FlushBuffers();
			}
			
			if(renderer != null){
				renderer.enabled = false;
			}
			
			if(!isPerson){
				foreach(var entry in originalRendererLayers){
					entry.Key.gameObject.layer = entry.Value;
				}	
			}
		}
		
		private void ShowMaterialOptionUI(int index){
			MaterialConfig matConfig = materialConfigs[index];
			if(!matConfig.enabled.val)
				return;
			
			matConfig.matOptionsUI.gameObject.SetActive(true);
			scriptUI.leftUIContent.gameObject.SetActive(false);
			scriptUI.fullWidthUIContent.gameObject.SetActive(true);
		}
		
		private void HideMaterialOptionUI(){
			foreach(MaterialConfig matConfig in materialConfigs){
				matConfig.matOptionsUI.gameObject.SetActive(false);
			}
			scriptUI.leftUIContent.gameObject.SetActive(true);
			scriptUI.fullWidthUIContent.gameObject.SetActive(false);
		}
		
		private void SyncMaterials(bool force = false){
			if(!force && oldMaterials != null && skin.GPUmaterials != oldMaterials){
				return;
			}
			
			Material[] mats = new Material[skin.GPUmaterials.Length];
			Material discard = new Material(Shader.Find("Custom/Discard"));
			
			// TODO: Add a no-cull shader option
			//Shader transparent = Shader.Find("Custom/Subsurface/TransparentGlossNMNoCullSeparateAlphaComputeBuff");
			//Shader transparent = Shader.Find("Custom/Subsurface/TransparentGlossNMSeparateAlphaComputeBuff");
			
			
			foreach(MaterialConfig matConfig in materialConfigs){
				if(!matConfig.enabled.val){
					mats[matConfig.index] = discard;
				} else {
					mats[matConfig.index] = matConfig.material;
				}
			}
			
			renderer.sharedMaterials = mats;
			renderer.enabled = true;
			skin.GPUmaterials = mats;
			skin.FlushBuffers();
		}
		
		private GameObject FindGameObject(GameObject parent, string name){
			Transform childTransform = parent.transform.Find(name);
			
			if(childTransform == null){
				return null;
			} else {
				return childTransform.gameObject;
			}
		}
		
		private GameObject GetOrCreateChildObject(GameObject parent, string name) {
			GameObject gameObject = FindGameObject(parent, name);
			if(gameObject == null) {
				gameObject = new GameObject(name);
				gameObject.transform.parent = parent.transform;
			} 
			
			return gameObject;
		}
		
		private Component GetOrCreateComponent(GameObject parent, System.Type type) {
			Component component = parent.GetComponent(type) as Component;
			if(component == null){
				component = parent.AddComponent(type) as Component;
			}
			return component;
		}
		
		List<GameObject> GetAllChildren(GameObject parent, List<GameObject> objectList = null)
		{
			if (objectList == null) objectList = new List<GameObject>();
		  
			foreach (Transform child in parent.transform) {
				objectList.Add(child.gameObject);
				GetAllChildren(child.gameObject, objectList);
			}
			return objectList;
		}
		
		public override void RestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, JSONArray presetAtoms = null, bool setMissingToDefault = true) {
			base.RestoreFromJSON(jc, restorePhysical, restoreAppearance, presetAtoms, setMissingToDefault);
			
			foreach(MaterialConfig matConfig in materialConfigs){
				JSONClass storedOpts = (JSONClass)jc[matConfig.name+"_opts"];
				
				matConfig.matOptions.RestoreFromJSON(storedOpts);
				matConfig.matOptions.SetParams();
			}
		}
		
		public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
		{
			var jsonClass = base.GetJSON(includePhysical, includeAppearance, forceStore);

			foreach(MaterialConfig matConfig in materialConfigs){
				JSONClass matOptsJson = matConfig.matOptions.GetJSON(true,true,true);
				
				jsonClass.Add(matConfig.name+"_opts", matOptsJson); 
			}
		
			return jsonClass;
		}
		
	}
}